// SPDX-License-Identifier: MIT
// Copyright (c) 2018-2024, Advanced Micro Devices, Inc. All rights reserved.

#pragma once
#include <stdint.h>

namespace ck_tile {

using index_t      = int32_t;
using long_index_t = int64_t;
using int8_t       = int8_t;

} // namespace ck_tile
